import sys
import os

sys.path = [os.path.dirname(os.path.realpath(__file__)) + "/../../src"] + sys.path

project = 'RiotKit Do'
copyright = '2019, RiotKit Collective'
author = 'RiotKit Collective'

version = ''
release = '1'

extensions = [
    'sphinx.ext.todo',
    'sphinx.ext.imgmath',
    'sphinx.ext.githubpages',
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinxcontrib.jinja'
]

jinja_contexts = {
    'py_publish': {
        'PKG_NAME': 'rkd_python',
        'PKG_CLASS_NAME': 'rkd_python.PublishTask',
        'PKG_PIP': 'rkd_python'
    },
    'py_build': {
        'PKG_NAME': 'rkd_python',
        'PKG_CLASS_NAME': 'rkd_python.BuildTask',
        'PKG_PIP': 'rkd_python'
    },
    'py_install': {
        'PKG_NAME': 'rkd_python',
        'PKG_CLASS_NAME': 'rkd_python.InstallTask',
        'PKG_PIP': 'rkd_python'
    },
    'py_clean': {
        'PKG_NAME': 'rkd_python',
        'PKG_CLASS_NAME': 'rkd_python.CleanTask',
        'PKG_PIP': 'rkd_python'
    },
    'py_unittest': {
        'PKG_NAME': 'rkd_python',
        'PKG_CLASS_NAME': 'rkd_python.UnitTestTask',
        'PKG_PIP': 'rkd_python'
    },
    'shell': {
        'PKG_NAME': 'rkd.standardlib.shell',
        'PKG_CLASS_NAME': 'rkd.standardlib.shell.ShellCommandTask',
        'PKG_PIP': 'rkd'
    },
    'exec': {
        'PKG_NAME': 'rkd.standardlib.shell',
        'PKG_CLASS_NAME': 'rkd.standardlib.shell.ExecProcessCommand',
        'PKG_PIP': 'rkd'
    },
    'j2_render': {
        'PKG_NAME': 'rkd.standardlib.jinja',
        'PKG_CLASS_NAME': 'rkd.standardlib.jinja.FileRendererTask',
        'PKG_PIP': 'rkd'
    },
    'j2_directory_to_directory': {
        'PKG_NAME': 'rkd.standardlib.jinja',
        'PKG_CLASS_NAME': 'rkd.standardlib.jinja.RenderDirectoryTask',
        'PKG_PIP': 'rkd'
    },
    'init': {
        'PKG_NAME': 'rkd.standardlib',
        'PKG_CLASS_NAME': 'rkd.standardlib.InitTask',
        'PKG_PIP': 'rkd'
    },
    'tasks': {
        'PKG_NAME': 'rkd.standardlib',
        'PKG_CLASS_NAME': 'rkd.standardlib.TasksListingTask',
        'PKG_PIP': 'rkd'
    },
    'callable_task': {
        'PKG_NAME': 'rkd.standardlib',
        'PKG_CLASS_NAME': 'rkd.standardlib.CallableTask',
        'PKG_PIP': 'rkd'
    },
    'rkd_create_structure': {
        'PKG_NAME': 'rkd.standardlib',
        'PKG_CLASS_NAME': 'rkd.standardlib.CreateStructureTask',
        'PKG_PIP': 'rkd'
    },
    'version': {
        'PKG_NAME': 'rkd.standardlib',
        'PKG_CLASS_NAME': 'rkd.standardlib.VersionTask',
        'PKG_PIP': 'rkd'
    },
    'env_get': {
        'PKG_NAME': 'rkd.standardlib.env',
        'PKG_CLASS_NAME': 'rkd.standardlib.env.SetEnvTask',
        'PKG_PIP': 'rkd'
    },
    'env_set': {
        'PKG_NAME': 'rkd.standardlib.env',
        'PKG_CLASS_NAME': 'rkd.standardlib.env.GetEnvTask',
        'PKG_PIP': 'rkd'
    },
    'line_in_file': {
        'PKG_NAME': 'rkd.standardlib',
        'PKG_CLASS_NAME': 'rkd.standardlib.LineInFileTask',
        'PKG_PIP': 'rkd'
    }
}

autodoc_inherit_docstrings = True
autodoc_default_flags = 'show-inheritance'

jinja_base = os.path.realpath(os.path.dirname(os.path.realpath(__file__)) + '/../')

templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
language = None
exclude_patterns = []
pygments_style = None

import sphinx_glpi_theme
html_theme = 'glpi'
html_theme_path = sphinx_glpi_theme.get_html_themes_path()

html_theme_options = {
    'body_max_width': None
}

html_css_files = [
    'css/riotkit.css',
]

html_static_path = ['_static']
htmlhelp_basename = 'RiotkitDoDoc'


latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'FileRepository.tex', 'RiotKit Do Documentation',
     'Wolnosciowiec Team', 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'filerepository', 'RiotKit Do Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'FileRepository', 'RiotKit Do Documentation',
     author, 'FileRepository', 'One line description of project.',
     'Miscellaneous'),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# -- Extension configuration -------------------------------------------------

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True



#!/usr/bin/env python3

import unittest
import os
import tempfile
from rkd.yaml_parser import YamlFileLoader
from rkd.exception import YAMLFileValidationError

SCRIPT_DIR_PATH = os.path.dirname(os.path.realpath(__file__))


class TestLoader(unittest.TestCase):
    def test_validates_successfully(self):
        yaml_loader = YamlFileLoader([])

        parsed = yaml_loader.load('''
            version: org.riotkit.rkd/yaml/v1
            imports: []
            tasks: {}
            ''', schema_name='org.riotkit.rkd/yaml/v1')

        self.assertIn('version', parsed)

    def test_raises_error_when_type_does_not_match(self):
        """Expect OBJECT at .tasks path, but get ARRAY instead"""

        yaml_loader = YamlFileLoader([])

        self.assertRaises(
            YAMLFileValidationError,
            lambda: yaml_loader.load('''
                version: org.riotkit.rkd/yaml/v1
                imports: []
                tasks: []
                ''', schema_name='org.riotkit.rkd/yaml/v1')
                          )

    def test_expect_path_will_be_shown_in_exception_message(self):
        """Simply check if path to the attribute will be printed within the exception"""

        yaml_loader = YamlFileLoader([])

        try:
            yaml_loader.load('''
                version: org.riotkit.rkd/yaml/v1
                imports: []
                tasks: 
                    :join:iwa-ait: []
                ''', schema_name='org.riotkit.rkd/yaml/v1')
        except YAMLFileValidationError as e:
            self.assertIn(
                "YAML schema validation failed at path \"tasks.:join:iwa-ait\" with error: [] is not of type 'object'",
                str(e)
            )
            return

        self.fail('Expected an exception to be raised')

    def test_expect_deeper_validation_will_be_performed(self):
        """Expects that argparse arguments will be validated"""

        yaml_loader = YamlFileLoader([])

        try:
            yaml_loader.load('''
version: org.riotkit.rkd/yaml/v1
imports: []
tasks: 
    :join:iwa-ait:
        description: Subscribe to any local section of IWA-AIT, workers have common interest
        arguments:
            - not a list
                        ''', schema_name='org.riotkit.rkd/yaml/v1')
        except YAMLFileValidationError as e:
            self.assertIn("tasks.:join:iwa-ait.arguments", str(e))
            self.assertIn("is not of type 'object'", str(e))
            return

        self.fail('Expected an exception to be raised')

    def test_loads_from_file_is_searching_in_rkd_path(self):
        """Assert that makefile.yml will be searched in RKD_PATH"""

        yaml_loader = YamlFileLoader([])

        d = tempfile.TemporaryDirectory()
        os.environ['RKD_PATH'] = d.name

        with open(d.name + '/makefile.yml', 'w') as f:
            f.write('''
version: org.riotkit.rkd/yaml/v1
imports: []
tasks: 
    :join:iwa-ait:
        description: Subscribe to any local section of IWA-AIT, workers have common interest
        arguments:
            - not a list
            ''')

        try:
            self.assertRaises(YAMLFileValidationError,
                              lambda: yaml_loader.load_from_file('makefile.yml', 'org.riotkit.rkd/yaml/v1'))
        finally:
            d.cleanup()
            os.environ['RKD_PATH'] = ''

    def test_invalid_file_path_is_causing_exception(self):
        """Test that invalid path will be reported quickly"""

        yaml_loader = YamlFileLoader([])
        self.assertRaises(FileNotFoundError,
                          lambda: yaml_loader.load_from_file('non-existing-file.yml', 'org.riotkit.rkd/yaml/v1'))

    def test_get_lookup_paths_includes_internal_path_as_well_as_rkd_path(self):
        """Verify that lookup paths includes RKD_PATH and internal RKD directories"""

        yaml_loader = YamlFileLoader([])
        os.environ['RKD_PATH'] = 'SOME-PATH-THERE'

        try:
            paths = yaml_loader.get_lookup_paths('harbor-internal/')
        finally:
            os.environ['RKD_PATH'] = ''

        defined_by_rkd_path = paths.index('SOME-PATH-THERE/harbor-internal/')
        internal_path = paths.index(os.path.realpath(SCRIPT_DIR_PATH + '/../src') + '/harbor-internal/')

        self.assertGreater(defined_by_rkd_path, internal_path, msg='defined_by_rkd_path should be favored')

    def test_find_path_by_name_founds_path(self):
        """Assert that makefile.yml will be searched in RKD_PATH"""

        yaml_loader = YamlFileLoader([])

        d = tempfile.TemporaryDirectory()
        os.environ['RKD_PATH'] = d.name

        with open(d.name + '/makefile.yml', 'w') as f:
            f.write('''
        version: org.riotkit.rkd/yaml/v1
        imports: []
        tasks: 
            :join:iwa-ait:
                description: Subscribe to any local section of IWA-AIT, workers have common interest
                arguments:
                    - not a list
                    ''')

        try:
            path = yaml_loader.find_path_by_name('makefile.yml', '/')
            self.assertTrue(len(path) > 0)
        finally:
            d.cleanup()
            os.environ['RKD_PATH'] = ''

    def test_find_path_by_name_does_not_found_anything(self):
        """Verify that find_path_by_name() will not return anything if nothing searched was found"""

        yaml_loader = YamlFileLoader([])
        self.assertEqual('', yaml_loader.find_path_by_name('some-file-that-does-not-exists', ''))



#!/usr/bin/env python3

from typing import List
from typing import Tuple
from argparse import ArgumentParser
from argparse import RawTextHelpFormatter
from .api.contract import TaskDeclarationInterface
from .api.contract import ArgumentEnv


class TraceableArgumentParser(ArgumentParser):
    """Traces options inserted into ArgumentParser for later interpretation

    Example case: ArgumentParser does not allow to check what is the default defined value for argument
                  but we need this information somewhere, so we track add_argument() parameters
                  (it is safe as the API is stable)
    """

    traced_arguments: dict

    def __init__(self, *args, **kwargs):
        self.traced_arguments = {}
        super().__init__(*args, **kwargs)

    def add_argument(self, *args, **kwargs):
        self._trace_argument(args, kwargs)
        super().add_argument(*args, **kwargs)

    def _trace_argument(self, args: tuple, kwargs: dict):
        for arg in args:
            self.traced_arguments[arg] = {
                'default': kwargs['default'] if 'default' in kwargs else None
            }

    def get_traced_argument(self, name: str):
        return self.traced_arguments[name]


class TaskArguments(object):
    _name: str
    _args: list

    def __init__(self, task_name: str, args: list):
        self._name = task_name
        self._args = args

    def __repr__(self):
        return 'Task<%s (%s)>' % (self._name, str(self._args))

    def name(self):
        return self._name

    def args(self):
        return self._args


class CommandlineParsingHelper(object):
    """
    Extends argparse functionality by grouping arguments into tasks -> tasks arguments
    """

    @classmethod
    def create_grouped_arguments(cls, commandline: List[str]) -> List[TaskArguments]:
        current_group_elements = []
        current_task_name = 'rkd:initialize'
        tasks = []
        cursor = -1
        max_cursor = len(commandline)

        for part in commandline:
            cursor += 1

            # normalize - strip out spaces to be able to detect "-", "--" and ":" at the beginning of string
            part = part.strip()

            is_flag = part[0:1] == "-"
            is_task = part[0:1] in (':', '@')
            previous_is_flag = commandline[cursor-1][0:1] == "-" if cursor >= 1 else False

            # option name or flag
            if is_flag:
                current_group_elements.append(part)

            # option value
            elif not is_flag and previous_is_flag and not is_task:
                current_group_elements.append(part)

            # new task
            elif is_task:
                if current_task_name != 'rkd:initialize':
                    tasks.append([current_task_name, current_group_elements])

                current_task_name = part
                current_group_elements = []

            # is not an option (--some or -s) but an argument actually
            else:
                current_group_elements.append(part)

            if cursor + 1 == max_cursor:
                tasks.append([current_task_name, current_group_elements])

        return cls._map_to_task_arguments(
            cls._parse_shared_arguments(tasks)
        )

    @classmethod
    def _map_to_task_arguments(cls, tasks: list) -> List[TaskArguments]:
        return list(map(
            lambda task: TaskArguments(task[0], task[1]),
            tasks
        ))

    @classmethod
    def _parse_shared_arguments(cls, tasks: list) -> list:
        """Apply arguments from task "@" that is before a group of tasks
           "@" without any arguments is clearing previous "@" with arguments
        """

        global_group_elements = []
        edited_tasks = []

        for task in tasks:
            task_name, group_elements = task

            if task_name == '@':
                global_group_elements = group_elements
                continue

            edited_tasks.append([
                task_name,
                group_elements + global_group_elements
            ])

        return edited_tasks

    @classmethod
    def parse(cls, task: TaskDeclarationInterface, args: list) -> Tuple[dict, dict]:
        """Parses ArgumentParser arguments defined by tasks

        Behavior:
          - Adds RKD-specific arguments
          - Includes task's specific arguments
          - Formats description, including documentation of environment variables

        Returns:
          Tuple of two dicts. First dict: arguments key=>value, Second dict: arguments definitions for advanced usae
        """

        argparse = TraceableArgumentParser(task.to_full_name(), formatter_class=RawTextHelpFormatter)

        argparse.add_argument('--log-to-file', '-rf', help='Capture stdout and stderr to file')
        argparse.add_argument('--log-level', '-rl', help='Log level: debug,info,warning,error,fatal')
        argparse.add_argument('--keep-going', '-rk', help='Allow going to next task, even if this one fails',
                              action='store_true')
        argparse.add_argument('--silent', '-rs', help='Do not print logs, just task output', action='store_true')
        argparse.add_argument('--become', '-rb', help='Execute task as given user (requires sudo)', default='')

        task.get_task_to_execute().configure_argparse(argparse)
        cls.add_env_variables_to_argparse(argparse, task)

        return vars(argparse.parse_args(args)), argparse.traced_arguments

    @classmethod
    def add_env_variables_to_argparse(cls, argparse: ArgumentParser, task: TaskDeclarationInterface):
        if argparse.description is None:
            argparse.description = ""

        argparse.description += task.get_full_description() + "\n"

        # print all environment variables possible to use
        argparse.description += "\nEnvironment variables for task \"%s\":\n" % task.to_full_name()

        for env in task.get_task_to_execute().internal_normalized_get_declared_envs().values():
            env: ArgumentEnv

            argparse.description += " - %s (default: %s)\n" % (
                str(env.name), str(env.default)
            )

        if not task.get_task_to_execute().get_declared_envs():
            argparse.description += ' -- No environment variables declared -- '



#!/usr/bin/env python3

import os
import sys
import subprocess
from io import FileIO
from typing import Optional
from threading import Thread

ON_POSIX = 'posix' in sys.builtin_module_names


class TextBuffer(object):
    text: str
    size: int

    def __init__(self, buffer_size: int):
        self.text = ''
        self.size = buffer_size

    def write(self, text: str):
        buf_len = len(self.text)
        txt_len = len(text)

        if buf_len + txt_len > self.size:
            missing_len = (buf_len + txt_len) - self.size
            self.trim_left_by(missing_len)

        self.text += text

    def trim_left_by(self, bytes: int):
        self.text = self.text[bytes:]

    def get_value(self) -> str:
        return self.text


def check_call(command: str, stdin=None, script: Optional[str] = ''):
    if os.getenv('RKD_COMPAT_SUBPROCESS') == 'true':
        subprocess.check_call(command, stdin=stdin, shell=True)
        return

    os.environ['PYTHONUNBUFFERED'] = "1"

    process = subprocess.Popen(command, shell=True, stdin=stdin, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                               bufsize=1, close_fds=ON_POSIX, universal_newlines=True)

    out_buffer = TextBuffer(buffer_size=1024 * 10)
    stdout_thread = Thread(target=push_output, args=(process.stdout, sys.stdout, out_buffer))
    stdout_thread.daemon = True
    stdout_thread.start()

    exit_code = process.wait()

    if exit_code > 0:
        raise subprocess.CalledProcessError(
            exit_code, script if script else command, stderr=out_buffer.get_value(), output=out_buffer.get_value()
        )


def push_output(input_stream: FileIO, output_stream, out_buffer: TextBuffer):
    for line in iter(input_stream.readline, ''):
        output_stream.write(line)
        output_stream.flush()
        out_buffer.write(line)

    input_stream.close()


# def _debug(msg):
#     with open('/dev/stdout', 'w') as f:
#         f.write(str(msg))




import re
import sys
import os
import subprocess
from traceback import format_exc as py_format_exception
from json import dumps as json_encode
from json import loads as json_decode
from copy import deepcopy
from time import sleep
from typing import List
from getpass import getpass
from contextlib import contextmanager
from datetime import datetime
from ..exception import InterruptExecution


this = sys.modules[__name__]
this.IS_CAPTURING_DESCRIPTORS = False

LEVEL_DEBUG = 37
LEVEL_INFO = 36
LEVEL_WARNING = 33
LEVEL_ERROR = 31
LEVEL_FATAL = 20

LOG_LEVELS = {
    'debug': LEVEL_DEBUG,
    'info': LEVEL_INFO,
    'warning': LEVEL_WARNING,
    'error': LEVEL_ERROR,
    'fatal': LEVEL_FATAL
}

LOG_LEVEL_FORMATTING_MAPPING = {
    'debug':   "\x1B[0m%TEXT%\x1B[0m",
    'info':    "\x1B[1m%TEXT%\x1B[0m",
    'warn':    "\x1B[93m%TEXT%\x1B[0m",
    'error':   "\x1B[91m%TEXT%\x1B[0m",
    'fatal':   "\x1B[91m\x1B[5m%TEXT%\x1B[0m"
}


class StandardOutputReplication(object):
    _out_streams: list

    def __init__(self, out_streams: list):
        self._out_streams = out_streams

    def write(self, buf):
        for stream in self._out_streams:
            try:
                stream.write(buf)
            except TypeError:
                try:
                    stream.write(buf.encode('utf-8'))
                except AttributeError:
                    stream.write(str(buf))

        self.flush()

    def fileno(self):
        return 1

    def flush(self):
        pass


class IO:
    """ Interacting with input and output - stdout/stderr/stdin, logging """

    silent = False
    log_level = LEVEL_INFO

    @contextmanager
    def capture_descriptors(self, target_files: List[str] = None, stream=None, enable_standard_out: bool = True):
        """Capture stdout and stderr from a block of code - use with 'with'"""

        if target_files is None:
            target_files = []

        if this.IS_CAPTURING_DESCRIPTORS:
            self.debug('Deep call to capture_descriptors()')

        this.IS_CAPTURING_DESCRIPTORS = True

        sys_stdout = sys.stdout
        sys_stderr = sys.stderr
        log_files = []

        outputs_stdout = []
        outputs_stderr = []

        # 1. Prepare standard out/err
        if enable_standard_out:
            outputs_stdout.append(sys_stdout)
            outputs_stderr.append(sys_stderr)

        # 2. Prepare logs
        for target_file in target_files:
            subprocess.call(['mkdir', '-p', os.path.dirname(target_file)])

            log_file = open(target_file, 'wb', buffering=0)
            log_file.no_flush = True
            log_files.append(log_file)

            outputs_stdout.append(log_file)
            outputs_stderr.append(log_file)

        # 3. Prepare StringIO
        if stream:
            outputs_stdout.append(stream)
            outputs_stderr.append(stream)

        # 4. Mock
        sys.stdout = StandardOutputReplication(outputs_stdout)
        sys.stderr = StandardOutputReplication(outputs_stderr)

        # 5. Action!
        yield

        # 6. Revert standard out/err
        sys.stdout = sys_stdout
        sys.stderr = sys.stderr

        # 7. Clean up: close all log files
        for log_file in log_files:
            log_file.close()

        this.IS_CAPTURING_DESCRIPTORS = False

    def inherit_silent(self, io: 'SystemIO'):
        self.silent = io.is_silent(consider_ui=False)

    def is_silent(self) -> bool:
        """Is output silent? In silent mode OPTIONAL MESSAGES are not shown"""

        return self.silent

    #
    # Log level - mutable setting
    #
    def set_log_level(self, desired_level_name: str):
        if desired_level_name not in LOG_LEVELS:
            raise Exception('Invalid log level name')

        self.log_level = LOG_LEVELS[desired_level_name]

    def is_log_level_at_least(self, log_level: str) -> bool:
        return self.log_level >= LOG_LEVELS[log_level]

    def get_log_level(self) -> str:
        for name, severity in LOG_LEVELS.items():
            if severity == self.log_level:
                return name

        raise Exception('Unset log level')

    #
    # Standard output/error
    #

    def _stdout(self, text):
        sys.stdout.write(text)

    def _stderr(self, text):
        sys.stderr.write(text)

    def out(self, text):
        """ Standard output """
        self._stdout(text)

    def outln(self, text):
        """ Standard output + newline """
        self.out(text)
        self._stdout("\n")

    def err(self, text):
        """ Standard error """
        self._stderr(text)

    def errln(self, text):
        """ Standard error + newline """
        self.err(text)
        self._stderr("\n")

    def opt_out(self, text):
        """ Optional output - fancy output skipped in --silent mode """

        if not self.is_silent():
            self.out(text)

    def opt_outln(self, text):
        """ Optional output - fancy output skipped in --silent mode + newline """

        if not self.is_silent():
            self.outln(text)

    #
    # Logs
    #

    def debug(self, text):
        """Logger: debug

        """
        if self.log_level >= LEVEL_DEBUG:
            self.log(text, 'debug')

    def info(self, text):
        """Logger: info

        """

        if self.log_level >= LEVEL_INFO:
            self.log(text, 'info')

    def warn(self, text):
        """Logger: warn

        """

        if self.log_level >= LEVEL_WARNING:
            self.log(text, 'warn')

    def error(self, text):
        """Logger: error

        """

        if self.log_level >= LEVEL_ERROR:
            self.log(text, 'error')

    def critical(self, text):
        """Logger: critical

        """

        if self.log_level >= LEVEL_FATAL:
            self.log(text, 'critical')

    def log(self, text, level: str):
        if not self.is_silent():
            current_time = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
            level = LOG_LEVEL_FORMATTING_MAPPING[level].replace('%TEXT%', level)

            self.outln("\x1B[2m[%s]\x1B[0m[%s]: \x1B[2m%s\x1B[0m" % (current_time, level, text))

    def print_group(self, text):
        """Prints a colored text inside brackets [text] (optional output)

        """

        self.opt_outln("\x1B[33m[%s]\x1B[0m" % text)

    #
    # Lines and separators
    #

    def print_line(self):
        """Prints a newline

        """

        self.outln('')

    def print_opt_line(self):
        """Prints a newline (optional output)
        """

        self.opt_outln('')

    def print_separator(self):
        """Prints a text separator (optional output)
        """
        self.opt_outln("\x1B[37m%s\x1B[0m" % '-----------------------------------')

    #
    #  Statuses
    #

    def success_msg(self, text):
        """Success message (optional output)
        """

        self.opt_outln("\x1B[92m%s\x1B[0m" % text)

    def error_msg(self, text):
        """Error message (optional output)
        """

        self.opt_outln("\x1B[91m%s\x1B[0m" % text)

    def info_msg(self, text):
        """Informational message (optional output)
        """

        self.opt_outln("\x1B[93m%s\x1B[0m" % text)

    #
    # Standard formatting
    #
    def h1(self, text):
        """Heading #1 (optional output)
        """

        self.opt_outln("\x1B[93m  ##> %s\x1B[0m" % text)

    def h2(self, text):
        """Heading #2 (optional output)
        """

        self.opt_outln("\x1B[93m   ===> %s\x1B[0m" % text)

    def h3(self, text):
        """Heading #3 (optional output)
        """

        self.opt_outln("\x1B[33m    --> %s\x1B[0m" % text)

    def h4(self, text):
        """Heading #3 (optional output)
        """

        self.opt_outln("\x1B[33m     ... %s\x1B[0m" % text)


class SystemIO(IO):
    """ Used for logging outside of tasks """

    _ui = True

    def capture_descriptors(self, target_file: str = None, stream=None, enable_standard_out: bool = True):
        pass

    def set_display_ui(self, ui: bool):
        self._ui = ui

    def is_silent(self, consider_ui: bool = True) -> bool:
        if consider_ui and not self._ui:
            return True

        return self.silent


class NullSystemIO(SystemIO):
    def _stdout(self, text):
        pass

    def _stderr(self, text):
        pass


class BufferedSystemIO(SystemIO):
    _buffer = ''

    def _stdout(self, text):
        self._buffer += text

    def _stderr(self, text):
        self._buffer += text

    def get_value(self):
        return self._buffer

    def clear_buffer(self):
        self._buffer = ''


class Wizard(object):
    _max_retries: int = 3
    answers: dict
    io: 'IO'
    task: 'TaskInterface'
    to_env: dict
    sleep_time = 1
    filename: str

    def __init__(self, task: 'TaskInterface', filename: str = 'tmp-wizard.json'):
        self.answers = {}
        self.task = task
        self.io = task.io()
        self.to_env = {}
        self.filename = filename

    def ask(self, title: str, attribute: str, regexp: str = '', to_env: bool = False, default: str = None,
            choices: list = [], secret: bool = False) -> 'Wizard':
        """Asks user a question

        Usage:
            wizard = Wizard(self)
            wizard.ask('In which year the Spanish social revolution has begun?',
                       attribute='year',
                       choices=['1936', '1910'])
            wizard.finish()

        """

        retried = 0
        value = None
        full_text_to_ask = title

        if choices and regexp:
            raise Exception('Please choose between regexp and choices validation.')

        if regexp:
            full_text_to_ask += " [%s]" % regexp

        if default:
            full_text_to_ask += " [default: %s]" % default

        if choices:
            full_text_to_ask += " [%s]" % ', '.join(choices)

        full_text_to_ask += ": "

        while value is None or not self.is_valid(value, regexp, choices):
            self.io.out(full_text_to_ask + "\n -> ")
            value = self.input(secret=secret)

            if default and not value.strip():
                value = default

            if retried >= self._max_retries:
                raise InterruptExecution('Invalid value given')

            if self.is_valid(value, regexp):
                break

            retried += 1
            sleep(self.sleep_time)

        if to_env:
            self.to_env[attribute] = value
            return self

        self.answers[attribute] = value
        return self

    def input(self, secret: bool = False):
        """Extracted for unit testing to be possible easier"""

        if os.getenv('__WIZARD_INPUT'):
            return os.getenv('__WIZARD_INPUT')

        if secret:
            return getpass(prompt='')

        return input()

    @staticmethod
    def is_valid(value: any, regexp: str = '', choices: list = []):
        if choices:
            return value in choices

        if regexp:
            return re.match(regexp, value) is not None

        return True

    def load_previously_stored_values(self):
        """Load previously saved values"""

        if os.path.isfile('.rkd/' + self.filename):
            with open('.rkd/' + self.filename, 'rb') as f:
                self.answers = json_decode(f.read())

        self.to_env = deepcopy(os.environ)

    def finish(self) -> 'Wizard':
        """Commit all pending changes into json and .env files"""

        self.io.info('Writing to .rkd/' + self.filename)
        with open('.rkd/' + self.filename, 'wb') as f:
            f.write(json_encode(self.answers).encode('utf-8'))

        self.io.info('Writing to .env')
        for attribute, value in self.to_env.items():
            self.task.rkd(
                [':env:set', '--name="%s"' % attribute, '--value="%s"' % value],
                verbose=False,
                capture=True
            )

        return self


def clear_formatting(text: str) -> str:
    text = re.sub("\\x1B\[([0-9]+)m", '', text)

    return text


def output_formatted_exception(exc: Exception, title: str, io: IO):
    """Formats a catched exception and displays as user-friendly by default (without a stack trace)

    When at least there is a "debug" level of error reporting, then an original stack trace would be displayed
    Everything goes through the RKD's IO, not directly to the stdout/stderr. The stderr is used there naturally.
    """

    try:
        io.errln('During "%s" a critical error happened' % title)
        io.print_line()

        if io.is_log_level_at_least('debug'):
            io.errln(py_format_exception())
            return

        cause = exc.__cause__

        while cause is not None:
            io.errln('\x1B[91m(Caused by) %s: \x1B[93m%s\x1B[0m' % (
                cause.__class__.__name__,
                indent_new_lines(str(cause), 4)
            ))

            cause = cause.__cause__

        io.errln('\x1B[91m%s: \x1B[93m%s\x1B[0m' % (exc.__class__.__name__, indent_new_lines(str(exc), 4)))
        io.print_line()
        io.errln('\x1B[37mRetry with "-rl debug" switch before failed task to see stacktrace\x1B[0m')

    except Exception:
        print('FATAL: During exception formatting there was an unrecoverable error')
        print(py_format_exception())
        sys.exit(1)

    return


def indent_new_lines(text: str, num: int = 4):
    """Inserts spaces at the beginning of each new line"""

    return text.replace("\n", "\n" + (" " * num))




"""
Declarative allows to collect a list of steps with BASH and/or Python code, then later execute it

Especially helpful, when parsing what is to execute - used in MAKEFILE YAML syntax

"""

import os
import ast
from collections import OrderedDict
from collections import namedtuple
from copy import deepcopy
from traceback import format_exc
from typing import List
from ..api.contract import TaskInterface
from ..api.contract import ExecutionContext


Step = namedtuple('Step', ['language', 'code', 'task_name', 'rkd_path', 'envs', 'task_num'])


class DeclarativeExecutor(object):
    """Executes declared Bash and/or Python code in form of steps

    Avoids using lambdas and inner-methods, so the code is simple, 100% statically typed and serializable
    """

    steps: List[Step]

    def __init__(self):
        self.steps = []

    def add_step(self, language: str, code: str, task_name: str, rkd_path: str, envs: dict):
        self.steps.append(Step(
            language=language,
            code=code, task_name=task_name,
            rkd_path=rkd_path,
            envs=envs,
            task_num=len(self.steps) + 1
        ))

    def execute_steps_one_by_one(self, ctx: ExecutionContext, this: TaskInterface) -> bool:
        """Proxy that executes all steps one-by-one as part of TaskInterface.execute()"""

        this.io().debug('Executing declared steps one-by-one')

        for step in self.steps:
            step: Step

            if step.language == 'python':
                execute = self.execute_python_step
            elif step.language == 'bash':
                execute = self.execute_bash_step
            else:
                raise Exception('Invalid language type')

            this.io().debug('Executing step %i' % step.task_num)
            result = execute(ctx, this, step)

            # if one of step failed, then interrupt and mark task as failure
            if not result:
                this.io().debug('Step failed, interrupting task execution')
                return False

        return True

    @staticmethod
    def execute_python_step(ctx: ExecutionContext, this: TaskInterface, step: Step) -> bool:
        """Executes a Python code through eval()

        The "ctx" and "this" variables are available internally as a context for executed code
        """

        env_backup = deepcopy(os.environ)

        # "ctx" and "this" will be available as a local context
        try:
            # prepare environment
            os.environ.update(step.envs)
            os.environ['RKD_PATH'] = step.rkd_path
            this.push_env_variables(os.environ)
            filename = step.task_name + '@step ' + str(step.task_num)

            # compile code
            this.io().debug('Compiling code')
            tree = ast.parse(step.code)

            if not isinstance(tree.body[-1], ast.Return):
                this.io().debug('Python code at step %s@%i does not have return' % (step.task_name, step.task_num))
                tree = ast.parse(step.code + "\nreturn False")

            eval_expr = ast.Expression(tree.body[-1].value)
            exec_expr = ast.Module(tree.body[:-1], type_ignores=[])

            # run compiled code
            exec(compile(exec_expr, filename, 'exec'))

            to_return = eval(compile(eval_expr, filename, 'eval'))

        except Exception as e:
            this.io().error_msg('Error while executing step %i in task "%s". Exception: %s' % (
                step.task_num, step.task_name, str(e)
            ))
            this.io().error_msg(format_exc())

            to_return = False

        finally:
            os.environ = env_backup

        return to_return

    @staticmethod
    def execute_bash_step(ctx: ExecutionContext, this: TaskInterface, step: Step) -> bool:
        try:
            # assign arguments from ArgumentParser (argparse) to env variables
            args = OrderedDict()
            for name, value in ctx.args.items():
                args['ARG_' + name.upper()] = value

            # glue all environment variables
            process_env = OrderedDict()
            process_env.update(step.envs)
            process_env.update(args)
            process_env.update({'RKD_PATH': step.rkd_path, 'RKD_DEPTH': int(os.getenv('RKD_DEPTH', 0))})

            this.sh(step.code, strict=True, env=process_env)
            return True

        except Exception as e:
            this.io().error_msg('Error while executing step %i in task "%s". Exception: %s' % (
                step.task_num, step.task_name, str(e)
            ))
            return False




import os
import re
from .api.syntax import TaskDeclaration
from .context import ApplicationContext


def decide_about_target_log_files(ctx: ApplicationContext, log_to_file: str, declaration: TaskDeclaration,
                                  task_num: int):
    """Decides where to save logs"""

    log_files = []
    session_log: bool = os.getenv('RKD_AUDIT_SESSION_LOG', '').lower() in ['true', '1', 'yes']

    if log_to_file:
        log_files.append(log_to_file)

    if session_log:
        date = ctx.get_creation_date()
        time_with_seconds = date.strftime('%H:%M:%S.%f').replace(':', '_').replace('.', '-')
        template = '.rkd/logs/%DATE-DAY%/%DATE-SECOND%/task-%TASKNUM%-%NORMALIZED_TASKNAME%.log'

        log_files.append(
            template.replace('%DATE-DAY%', date.strftime('%Y-%m-%d'))
                    .replace('%DATE-SECOND%', time_with_seconds)
                    .replace('%TASKNUM%', str(task_num))
                    .replace('%NORMALIZED_TASKNAME%', normalize_task_name_to_filename(declaration))
        )

    return log_files


def normalize_task_name_to_filename(declaration: TaskDeclaration) -> str:
    return re.sub('[^a-zA-Z0-9_]+', '_', declaration.to_full_name()[1:])




from typing import Union
from .api.syntax import TaskDeclaration, GroupDeclaration
from .argparsing import CommandlineParsingHelper
from .exception import NotSupportedEnvVariableError


class TaskDeclarationValidator(object):

    @staticmethod
    def assert_declaration_is_valid(task: TaskDeclaration, task_num: int,
                                    parent: Union[GroupDeclaration, None] = None,
                                    args: list = []):

        # check if arguments are satisfied
        CommandlineParsingHelper.parse(task, args)

        # validate environment variables
        allowed_envs = task.get_task_to_execute().internal_normalized_get_declared_envs()

        for env_name in task.get_user_overridden_envs():
            if env_name not in allowed_envs:
                raise NotSupportedEnvVariableError('"%s" is not a supported environment variable by "%s" task' % (
                    env_name, task.to_full_name()
                ))




import os
import pwd
from pickle import dumps as pickle_dumps
from pickle import loads as pickle_loads
from typing import Union
from ..argparsing import CommandlineParsingHelper
from ..api.syntax import TaskDeclaration, GroupDeclaration
from ..api.contract import TaskInterface
from ..api.contract import ExecutorInterface
from ..api.contract import ExecutionContext
from ..context import ApplicationContext
from ..inputoutput import IO
from ..inputoutput import SystemIO
from ..inputoutput import output_formatted_exception
from .results import ProgressObserver
from ..exception import InterruptExecution
from ..audit import decide_about_target_log_files
from ..api.temp import TempManager
from .serialization import FORKED_EXECUTOR_TEMPLATE
from .serialization import get_unpicklable


class OneByOneTaskExecutor(ExecutorInterface):
    """ Executes tasks one-by-one, providing a context that includes eg. parsed arguments """

    _ctx: ApplicationContext
    _observer: ProgressObserver
    io: SystemIO

    def __init__(self, ctx: ApplicationContext):
        self._ctx = ctx
        self.io = ctx.io
        self._observer = ProgressObserver(ctx.io)

    def execute(self, declaration: TaskDeclaration, task_num: int, parent: Union[GroupDeclaration, None] = None,
                args: list = []):

        """ Executes a single task passing the arguments, redirecting/capturing the output and handling the errors """

        result = False
        is_exception = False

        # 1. notify
        self._observer.task_started(declaration, parent, args)

        # 2. parse arguments
        parsed_args, defined_args = CommandlineParsingHelper.parse(declaration, args)
        log_level: str = parsed_args['log_level']
        log_to_file: str = parsed_args['log_to_file']
        is_silent: bool = parsed_args['silent']
        keep_going: bool = parsed_args['keep_going']
        cmdline_become: str = parsed_args['become']

        # 3. execute
        temp = TempManager()

        try:
            io = IO()
            io.set_log_level(log_level if log_level else self.io.get_log_level())

            if is_silent:
                io.silent = is_silent
            else:
                io.inherit_silent(self.io)  # fallback to system-wide

            where_to_store_logs = decide_about_target_log_files(self._ctx, log_to_file, declaration, task_num)

            with io.capture_descriptors(target_files=where_to_store_logs):

                task = declaration.get_task_to_execute()
                task.internal_inject_dependencies(io, self._ctx, self, temp)

                result = self._execute_directly_or_forked(cmdline_become, task, temp, ExecutionContext(
                        declaration=declaration,
                        parent=parent,
                        args=parsed_args,
                        env=declaration.get_env(),
                        defined_args=defined_args
                    ))

        # 4. capture result
        except Exception as e:
            # allows to keep going on, even if task fails
            if not keep_going:
                output_formatted_exception(e, str(task.get_full_name()), self.io)
                raise InterruptExecution()

            self._observer.task_errored(declaration, e)
            is_exception = True

        finally:
            temp.finally_clean_up()

            if result is True:
                self._observer.task_succeed(declaration, parent)
            else:
                if not is_exception:  # do not do double summary
                    self._observer.task_failed(declaration, parent)

                # break the whole pipeline only if not --keep-going
                if not keep_going:
                    raise InterruptExecution()

    def _execute_directly_or_forked(self, cmdline_become: str, task: TaskInterface, temp: TempManager, ctx: ExecutionContext):
        """Execute directly or pass to a forked process
        """

        if task.should_fork() or cmdline_become:
            task.io().debug('Executing task as separate process')
            return self._execute_as_forked_process(cmdline_become, task, temp, ctx)

        return task.execute(ctx)

    @staticmethod
    def _execute_as_forked_process(become: str, task: TaskInterface, temp: TempManager, ctx: ExecutionContext):
        """Execute task code as a separate Python process

        The communication between processes is with serialized data and text files.
        One text file is a script, the task code is passed with stdin together with a whole context
        Second text file is a return from executed task - it can be a boolean or exeception.

        When an exception is returned by a task, then it is reraised there - so the original exception is shown
        without any proxies.
        """

        if not become:
            become = task.get_become_as()

        # prepare file with source code and context
        communication_file = temp.assign_temporary_file()
        task.io().debug('Assigning communication temporary file at "%s"' % communication_file)

        context_to_pickle = {'task': task, 'ctx': ctx, 'communication_file': communication_file}

        try:
            task.io().debug('Serializing context')
            with open(communication_file, 'wb') as f:
                f.write(pickle_dumps(context_to_pickle))

        except (AttributeError, TypeError) as e:
            task.io().error('Cannot fork, serialization failed. ' +
                            'Hint: Tasks that are using internally inner-methods and ' +
                            'lambdas cannot be used with become/fork')
            task.io().error(str(e))

            if task.io().is_log_level_at_least('debug'):
                task.io().error('Pickle trace: ' + str(get_unpicklable(context_to_pickle)))

            return False

        code_file = temp.assign_temporary_file()
        task.io().debug('Assigned source code temporary file at "%s"' % code_file)

        with open(code_file, 'w') as f:
            f.write(FORKED_EXECUTOR_TEMPLATE)

        # set permissions to temporary file
        if become:
            task.io().debug('Setting temporary file permissions')
            os.chmod(communication_file, 0o777)

            try:
                pwd.getpwnam(become)
            except KeyError:
                task.io().error('Unknown user "%s"' % become)
                return False

        task.io().debug('Executing python code')
        task.py(communication_file, become=become, capture=False, script_path=code_file)

        # collect, process and pass result
        task.io().debug('Parsing subprocess results from a serialized data')
        with open(communication_file, 'rb') as conn_file:
            task_return = pickle_loads(conn_file.read())

        if isinstance(task_return, Exception):
            task.io().debug('Exception was raised in subprocess, re-raising')
            raise task_return

        return task_return

    def get_observer(self) -> ProgressObserver:
        return self._observer



#!/usr/bin/env python3

import unittest
import os
from io import StringIO
from rkd.contract import ExecutionContext
from rkd.execution.executor import OneByOneTaskExecutor
from rkd.context import ApplicationContext
from rkd.test import get_test_declaration
from rkd.test import ret_true
from rkd.test import ret_invalid_user
from rkd.api.temp import TempManager
from rkd.api.inputoutput import BufferedSystemIO
from rkd.api.inputoutput import IO
from rkd.contract import TaskInterface
from rkd.test import TestTask
from rkd.test import TestTaskWithRKDCallInside

CURRENT_SCRIPT_PATH = os.path.dirname(os.path.realpath(__file__))


class TestOneByOneExecutor(unittest.TestCase):
    @staticmethod
    def _prepare_test_for_forking_process(task: TaskInterface = None):
        if not task:
            task = TestTask()

        io = IO()
        string_io = StringIO()

        temp = TempManager(chdir='/tmp/')
        container = ApplicationContext([], [], '')
        container.io = BufferedSystemIO()
        executor = OneByOneTaskExecutor(container)

        declaration = get_test_declaration(task)
        task._io = io
        task._io.set_log_level('debug')
        ctx = ExecutionContext(declaration)

        return string_io, task, executor, io, ctx, temp

    def test_execute_directly_or_forked_asks_task_for_forking(self):
        """Check that decision of forking or not forking belongs to the Task"""

        string_io, task, executor, io, ctx, temp = self._prepare_test_for_forking_process()

        # mock to get results instead of real action
        executor._execute_as_forked_process = lambda *args, **kwargs: expectations.append('executor::_execute_as_forked_process')
        task.execute = lambda *args, **kwargs: expectations.append('task::execute')

        with self.subTest('Will fork'):
            expectations = []
            task.should_fork = lambda: True
            executor._execute_directly_or_forked('', task, temp, ctx)

            self.assertEqual(['executor::_execute_as_forked_process'], expectations)

        with self.subTest('Will not fork'):
            expectations = []
            task.should_fork = lambda: False
            executor._execute_directly_or_forked('', task, temp, ctx)

            self.assertEqual(['task::execute'], expectations)

    def test_execute_as_forked_process_executes_tasks_listing_task_in_a_separate_python_process(self):
        """Simply, successful case - execute a Hello World task code in a separate process and capture output"""

        string_io, task, executor, io, ctx, temp = self._prepare_test_for_forking_process()

        # mock
        task.should_fork = ret_true

        with io.capture_descriptors(stream=string_io, enable_standard_out=False):
            executor._execute_as_forked_process('', task, temp, ctx)

        self.assertIn('Hello world from :test task', string_io.getvalue())

    def test_execute_as_forked_process_will_inform_about_invalid_user(self):
        """Test that executing as forked process will abort if user does not exist in system"""

        string_io, task, executor, io, ctx, temp = self._prepare_test_for_forking_process()

        # mock
        task.get_become_as = ret_invalid_user

        with io.capture_descriptors(stream=string_io, enable_standard_out=False):
            executor._execute_as_forked_process('', task, temp, ctx)

        self.assertIn('Unknown user "invalid-user-there"', string_io.getvalue())

    def test_execute_as_forked_process_will_inform_about_unserializable_context(self):
        """Verify that tasks not serializable by pickle will be enough described with a clue,
        so the developer can know what is to correct (eg. a lambda returned as a method return)
        """

        string_io, task, executor, io, ctx, temp = self._prepare_test_for_forking_process()

        # mock: PUT NOT-SERIALIZABLE CALLABLE (LAMBDA)
        task.should_fork = lambda: True

        with io.capture_descriptors(stream=string_io, enable_standard_out=False):
            executor._execute_as_forked_process('', task, temp, ctx)

        self.assertIn('Pickle trace: ["[val type=TestTask].should_fork', string_io.getvalue())
        self.assertIn('Cannot fork, serialization failed. Hint: Tasks that are using internally' +
                      ' inner-methods and lambdas cannot be used with become/fork', string_io.getvalue())

    def test_execute_as_forked_process_will_not_break_rkd_method_inside_task_so_the_interpreter_path_will_be_properly_detected(self):
        """The forked process needs to know how to detect Python binary when using rkd()"""

        #
        # TestTaskWithRKDCallInside() contains a rkd() call inside execute(), that's what we are testing
        # it is invisible in our test. We cannot use lambda to define this in test, because lambdas are not serializable
        #
        string_io, task, executor, io, ctx, temp = self._prepare_test_for_forking_process(TestTaskWithRKDCallInside())
        task.should_fork = ret_true

        with io.capture_descriptors(stream=string_io, enable_standard_out=False):
            executor._execute_as_forked_process('', task, temp, ctx)

        self.assertIn('9 Aug 2014 Michael Brown, an unarmed Black teenager, ' +
                      'was killed by a white police officer in Ferguson, Missouri, ' +
                      'sparking mass protests across the US.',
                      string_io.getvalue())



#!/usr/bin/env python3

import unittest
import os
from rkd.test import TestTask
from rkd.contract import ArgumentEnv

CURRENT_SCRIPT_PATH = os.path.dirname(os.path.realpath(__file__))


class TestTaskInterface(unittest.TestCase):
    def test_table(self):
        """Simply test table() - the table is expected to use an external library, it is expected that external library
        will be tested already, but we need to check there if the interface matches
        """

        task = TestTask()
        out = task.table(
            header=['Activist', 'Born date'],
            body=[
                ['Mikhail Alexandrovich Bakunin', '1814'],
                ['Errico Malatesta', '1853'],
                ['José Buenaventura Durruti Dumange', '1896'],
                ['Johann Rudolf Rocker', '1873']
            ]
        )

        self.assertIn('---------------------------------', out)
        self.assertIn('Mikhail Alexandrovich Bakunin', out)
        self.assertIn('Johann Rudolf Rocker', out)
        self.assertIn('Activist', out)
        self.assertIn('Born date', out)

    def test_should_fork(self):
        task = TestTask()

        with self.subTest('Will fork'):
            task.get_become_as = lambda: 'root'
            self.assertTrue(task.should_fork())

        with self.subTest('Will not fork - no user specified'):
            task.get_become_as = lambda: ''
            self.assertFalse(task.should_fork())

    def test_internal_normalized_get_declared_envs_maps_primitive_types_into_class_instances(self):
        task = TestTask()
        task.get_declared_envs = lambda: {
            'SOME_ENV': 'primitive',
            'SOME_OTHER_ENV': ArgumentEnv(name='SOME_OTHER_ENV', switch='--cmd', default='not primitive')
        }

        normalized = task.internal_normalized_get_declared_envs()

        with self.subTest('Verify converted string -> ArgumentEnv'):
            self.assertTrue(isinstance(normalized['SOME_ENV'], ArgumentEnv))
            self.assertEqual('SOME_ENV', normalized['SOME_ENV'].name)
            self.assertEqual('--some-env', normalized['SOME_ENV'].switch)
            self.assertEqual('primitive', normalized['SOME_ENV'].default)

        with self.subTest('Verify not converted'):
            self.assertTrue(isinstance(normalized['SOME_OTHER_ENV'], ArgumentEnv))

    def test_internal_getenv_finds_mapped_environment_variable_by_switch_name(self):
        task = TestTask()
        task.get_declared_envs = lambda: {
            'SOME_OTHER_ENV': ArgumentEnv(name='SOME_OTHER_ENV', switch='--cmd', default='not primitive')
        }

        self.assertEqual('not primitive', task.internal_getenv('', envs={}, switch='--cmd'))

    def test_internal_getenv_finds_envronment_variable_by_its_not_mapped_name(self):
        task = TestTask()
        task.get_declared_envs = lambda: {
            'SOME_ENV': 'primitive'
        }

        with self.subTest('Should find by first argument SOME_ENV'):
            self.assertEqual('primitive', task.internal_getenv('SOME_ENV', envs={}, switch=''))

        with self.subTest('Should find by first argument, even if will not find for valid switch'):
            self.assertEqual('primitive', task.internal_getenv('SOME_ENV', envs={}, switch='--some-env'))

        with self.subTest('Should find by first argument, even if will not find for invalid switch'):
            self.assertEqual('primitive', task.internal_getenv('SOME_ENV', envs={}, switch='--some-non-existing'))





import os
from typing import Optional
from argparse import ArgumentParser
from rkd.api.contract import TaskInterface
from rkd.api.contract import ExecutionContext


class GetEnvTask(TaskInterface):
    """Gets environment variable value"""

    def get_name(self) -> str:
        return ':get'

    def get_group_name(self) -> str:
        return ':env'

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--name', '-e', help='Environment variable name', required=True)

    def execute(self, context: ExecutionContext) -> bool:
        # @todo: test for case, when None then ''
        self.io().out(os.getenv(context.get_arg('--name'), ''))

        return True


class SetEnvTask(TaskInterface):
    """Sets environment variable in the .env file"""

    def get_name(self) -> str:
        return ':set'

    def get_group_name(self) -> str:
        return ':env'

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--name', '-e', help='Environment variable name', required=True)
        parser.add_argument('--value', '-w', help='New value to set', required=False, default='')
        parser.add_argument('--ask', help='Ask for a value', required=False, action='store_true')
        parser.add_argument('--ask-text', help='Text to show, when asking for input', required=False, default='')

    def execute(self, context: ExecutionContext) -> bool:
        new_env_contents = ''
        env_name = context.get_arg('--name')
        new_value = context.get_arg('--value')
        path = os.getcwd() + '/.env'

        if context.get_arg('--ask'):
            txt = context.get_arg('--ask-text') if context.get_arg('--ask-text') else 'Please provide value for %s:' %\
                env_name
            new_value = input(txt + ' ')

        # write to existing file
        if os.path.isfile(path):
            wrote = False

            with open(path, 'r') as fo:
                for line in fo.readlines():
                    current_env_name = self.parse_env_name_from_line(line)

                    if current_env_name == env_name:
                        wrote = True
                        line = env_name + '=' + new_value + "\n"

                    new_env_contents += line

            if not wrote:
                new_env_contents += env_name + '=' + new_value + "\n"

        # create a new file
        else:
            new_env_contents += env_name + '=' + new_value + "\n"

        with open(path, 'w') as wo:
            wo.write(new_env_contents)

        return True

    @staticmethod
    def parse_env_name_from_line(line: str) -> Optional[str]:
        if "=" not in line:
            return None

        parts = line.split('=')

        # it may be a comment for example
        if ' ' in parts[0] or parts[0][0] == '#':
            return None

        return parts[0]



#!/usr/bin/env python3

import os
import unittest
from rkd.standardlib.jinja import RenderDirectoryTask
from rkd.test import mock_task, mock_execution_context

TESTS_DIR = os.path.dirname(os.path.realpath(__file__))


class TestRenderDirectoryTask(unittest.TestCase):
    """Tests for a task that should render JINJA2 files from DIRECTORY "A" to DIRECTORY "B"
    """

    @staticmethod
    def _execute_mocked_task(params: dict, env: dict = {}) -> tuple:
        task: RenderDirectoryTask = RenderDirectoryTask()
        mock_task(task)

        calls = []
        deletions = []

        def mock__render(source_path: str, target_path: str) -> bool:
            calls.append('"%s" -> "%s"' % (source_path, target_path))

            return True

        def mock__sh(*args, **kwargs):
            calls.append('sh(' + ' '.join(args) + ')')

        task._render = mock__render
        task.sh = mock__sh
        task._delete_file = lambda file: deletions.append(file)

        # run task
        task.execute(mock_execution_context(task, params, env))

        return calls, deletions

    def test_naming(self):
        self.assertEqual(':j2:directory-to-directory', RenderDirectoryTask().get_full_name())

    def test_iterates_over_subdirectories_including_depth_and_pattern(self):
        """Test with pattern, no source files deletion
        """
        renderings, deletions = self._execute_mocked_task({
            'source': '../',
            'target': '/tmp',
            'delete_source_files': False,
            'pattern': '(.*)(src|test)/(.*).py$',
            '--exclude-pattern': '',
            '--copy-not-matching-files': False,
            '--template-filenames': False
        })

        # example files (please correct if changed in filesystem)
        self.assertIn('"../test/test_standardlib_jinja_render_directory.py" ' +
                      '-> "/tmp/test/test_standardlib_jinja_render_directory.py"', renderings)
        self.assertIn('"../src/__init__.py" -> "/tmp/src/__init__.py"', renderings)

        # directories should not be included
        self.assertNotIn('"../src/" -> "/tmp/src/"', renderings)
        self.assertNotIn('"../src" -> "/tmp/src"', renderings)

    def test_files_are_copied_when_not_matching_pattern_but_switch_was_used(self):
        """Test --copy-not-matching-files switch that adds a possibility to copy all files from SOURCE to DESTINATION
        The difference is that those files that does not match PATTERN will be copied without rendering.

        Additionally uses "--exclude-pattern" to exclude redundant files
        """

        renderings, deletions = self._execute_mocked_task({
            'source': '../test',
            'target': '/tmp',
            'delete_source_files': False,
            'pattern': '(.*).j2',
            '--exclude-pattern': '(.*).pyc',
            '--copy-not-matching-files': True,
            '--template-filenames': False
        })

        flatten_list_as_str = ' '.join(renderings)

        with self.subTest('Check --copy-not-matching-files - the non (.*).j2 files should be just copied ' +
                          'instead of rendered'):

            self.assertIn('sh(cp -p "../test/test_standardlib_jinja_render_directory.py" ' +
                          '"/tmp//test_standardlib_jinja_render_directory.py")', renderings)

        with self.subTest('Check --exclude-pattern'):
            self.assertNotIn('.pyc', flatten_list_as_str)

    def test_without_pattern(self):
        """What happens if we specify no pattern?
        """
        renderings, deletions = self._execute_mocked_task({
            'source': '../',
            'target': '/tmp',
            'delete_source_files': False,
            'pattern': '',
            '--exclude-pattern': '',
            '--copy-not-matching-files': False,
            '--template-filenames': False
        })

        # example files (please correct if changed in filesystem)
        self.assertIn('README.', str(renderings))
        self.assertIn('requirements.txt', str(renderings))
        self.assertIn('__init__.py', str(renderings))

    def test_no_files_deleted_when_option_disabled(self):
        renderings, deletions = self._execute_mocked_task({
            'source': '../',
            'target': '/tmp',
            'delete_source_files': False,
            'pattern': '',
            '--exclude-pattern': '',
            '--copy-not-matching-files': False,
            '--template-filenames': False
        })

        self.assertEqual([], deletions)

    def test_files_are_called_to_be_deleted(self):
        renderings, deletions = self._execute_mocked_task({
            'source': '../',
            'target': '/tmp',
            'delete_source_files': True,
            'pattern': '(.*)test_standardlib_jinja_render_directory.py$',
            '--exclude-pattern': '',
            '--copy-not-matching-files': False,
            '--template-filenames': False
        })

        self.assertEqual(['../test/test_standardlib_jinja_render_directory.py'], deletions)

    def test_replace_vars_in_filename(self):
        name = RenderDirectoryTask().replace_vars_in_filename({'Word': 'triumph'}, 'that-agony-is-your---Word--.txt')

        self.assertEqual('that-agony-is-your-triumph.txt', name)

    def test_replace_vars_in_filename_multiple_occurrences(self):
        name = RenderDirectoryTask()\
            .replace_vars_in_filename({'word': 'pueblo', 'word2': 'hijos'}, '--word2--_del_--word--(--word--_version)')

        self.assertEqual('hijos_del_pueblo(pueblo_version)', name)

    def test_filename_templating_when_switch_is_on(self):
        """Assert that variables in filenames are also replaced, not only in the content

        Condition: When the "--template-filenames" is used
        """

        renderings, deletions = self._execute_mocked_task({
            'source': TESTS_DIR + '/internal-samples/jinja2-filename-templating',
            'target': '/tmp',
            'delete_source_files': True,
            'pattern': '(.*).j2',
            '--exclude-pattern': '',
            '--copy-not-matching-files': True,
            '--template-filenames': True
        }, env={
            'song_name': 'hijos-del-pueblo',
            'song2_name': 'a-las-barricadas'
        })

        self.assertIn('-> "/tmp//lyrics-hijos-del-pueblo.txt"', ' '.join(renderings),
                      msg='Expected that the matching .j2 file would have changed name')
        self.assertIn('"/tmp//lyrics2-a-las-barricadas.txt"', ' '.join(renderings),
                      msg='Expected that the non-j2 file would have also changed name')

    def test_filename_templating_is_not_replacing_vars_when_switch_is_not_used(self):
        """Checks if the variable replacing in filenames can be turned off by not using
        "--template-filenames" switch"""

        renderings, deletions = self._execute_mocked_task({
            'source': TESTS_DIR + '/internal-samples/jinja2-filename-templating',
            'target': '/tmp',
            'delete_source_files': True,
            'pattern': '(.*).j2',
            '--exclude-pattern': '',
            '--copy-not-matching-files': False,
            '--template-filenames': False
        }, env={
            'song_name': 'hijos-del-pueblo'
        })

        self.assertIn('-> "/tmp//lyrics---song_name--.txt"', ' '.join(renderings))



#!/usr/bin/env python3

import unittest
from typing import Union
from rkd.context import ApplicationContext
from rkd.resolver import TaskResolver
from rkd.standardlib.shell import ShellCommandTask
from rkd.syntax import TaskDeclaration, GroupDeclaration, TaskAliasDeclaration
from rkd.argparsing import TaskArguments
from rkd.aliasgroups import parse_alias_groups_from_env


class TestResolver(unittest.TestCase):
    def test_resolves_two_same_type_tasks_group_into_regular_tasks(self):
        """
        Tests that we can make an alias :test that will execute two times :sh command, but with different parameters
        :return:
        """

        context = ApplicationContext(
            tasks=[TaskDeclaration(ShellCommandTask())],
            aliases=[
                TaskAliasDeclaration(':test', [':sh', '-c', 'uname -a', ':sh', '-c', 'ps aux'],
                                     description='Task for testing')
            ],
            directory=''
        )

        context.compile()

        result_tasks = []

        def assertion_callback(declaration: TaskDeclaration,
                               task_num: int,
                               parent: Union[GroupDeclaration, None] = None,
                               args: list = []):
            result_tasks.append(declaration.to_full_name() + ' ' + (' '.join(declaration.get_args())))

        resolver = TaskResolver(context, [])
        resolver.resolve(
            [TaskArguments(':test', ['--short'])],
            assertion_callback
        )

        self.assertEqual([':sh -c uname -a', ':sh -c ps aux'], result_tasks)

    def test_resoles_regular_task(self):
        """Checks that :sh resolution works fine"""

        context = ApplicationContext(
            tasks=[TaskDeclaration(ShellCommandTask())],
            aliases=[],
            directory=''
        )

        context.compile()

        result_tasks = []

        def assertion_callback(declaration: TaskDeclaration,
                               task_num: int,
                               parent: Union[GroupDeclaration, None] = None,
                               args: list = []):
            result_tasks.append(declaration.to_full_name())

        resolver = TaskResolver(context, [])
        resolver.resolve([TaskArguments(':sh', [])], assertion_callback)

        self.assertEqual([':sh'], result_tasks)

    def test_resolves_aliased_task(self):
        """Checks 'alias groups' feature about to resolve some group name to other group name

        Example:
            :bella-ciao:sh -> :sh
        """

        context = ApplicationContext(
            tasks=[TaskDeclaration(ShellCommandTask())],
            aliases=[],
            directory=''
        )
        context.compile()
        result_tasks = []

        def assertion_callback(declaration: TaskDeclaration,
                               task_num: int,
                               parent: Union[GroupDeclaration, None] = None,
                               args: list = []):
            result_tasks.append(declaration.to_full_name())

        resolver = TaskResolver(context, parse_alias_groups_from_env(':bella-ciao->'))
        resolver.resolve([TaskArguments(':bella-ciao:sh', [])], assertion_callback)

        self.assertEqual([':sh'], result_tasks)



#!/usr/bin/env python3

import unittest
from argparse import ArgumentParser
from rkd.argparsing import CommandlineParsingHelper
from rkd.test import get_test_declaration


class ArgParsingTest(unittest.TestCase):
    def test_creates_grouped_arguments_into_tasks__task_after_flag(self):
        """ Test parsing arguments """

        parsed = CommandlineParsingHelper.create_grouped_arguments([
            ':harbor:start', '--profile=test', '--fast-fail', ':status'
        ])

        self.assertEqual("[Task<:harbor:start (['--profile=test', '--fast-fail'])>, Task<:status ([])>]", str(parsed))

    def test_creates_grouped_arguments_into_tasks__no_task_defined_goes_to_rkd_initialization(self):
        parsed = CommandlineParsingHelper.create_grouped_arguments([
            '--help'
        ])

        self.assertEqual("[Task<rkd:initialize (['--help'])>]", str(parsed))

    def test_creates_grouped_arguments_into_tasks__tasks_only(self):
        parsed = CommandlineParsingHelper.create_grouped_arguments([
            ':harbor:start', ':harbor:status', ':harbor:stop'
        ])

        self.assertEqual("[Task<:harbor:start ([])>, Task<:harbor:status ([])>, Task<:harbor:stop ([])>]", str(parsed))

    def test_add_env_variables_to_argparse(self):
        parser = ArgumentParser(':test')
        task = get_test_declaration()

        CommandlineParsingHelper.add_env_variables_to_argparse(parser, task)
        self.assertIn('ORG_NAME (default: International Workers Association)', parser.description)

    def test_add_env_variables_to_argparse__no_envs(self):
        parser = ArgumentParser(':test')
        task = get_test_declaration()

        # empty the values
        task.get_task_to_execute().get_declared_envs = lambda: {}

        CommandlineParsingHelper.add_env_variables_to_argparse(parser, task)
        self.assertNotIn('ORG_NAME (default: International Workers Association)', parser.description)
        self.assertIn('-- No environment variables declared --', parser.description)

    def test_arguments_usage(self):
        """Check that arguments are recognized"""

        parsed = CommandlineParsingHelper.create_grouped_arguments([
            ':strike:start', 'now'
        ])

        self.assertEqual("[Task<:strike:start (['now'])>]", str(parsed))

    def test_arguments_usage_with_switch_before(self):
        """Check that arguments are recognized - variant with an additional switch"""

        parsed = CommandlineParsingHelper.create_grouped_arguments([
            ':strike:start', '--general', 'now'
        ])

        self.assertEqual("[Task<:strike:start (['--general', 'now'])>]", str(parsed))

    def test_global_arguments_are_shared_for_all_tasks(self):
        """When we define a "@" task, then it is not present as a task (removed from tasks list), but its arguments
        are inherited to next tasks
        """

        parsed = CommandlineParsingHelper.create_grouped_arguments([
            '@', '--grassroot',
            ':strike:start', '--general', 'now',
            ':picket:start', '--at', 'exploiters-shop'
        ])

        self.assertEqual(
            "[Task<:strike:start (['--general', 'now', '--grassroot'])>, " +
            "Task<:picket:start (['--at', 'exploiters-shop', '--grassroot'])>]",
            str(parsed)
        )

    def test_global_arguments_are_cleared_after_inserting_alone_at_symbol(self):
        """When we define a "@" task, then it is not present as a task (removed from tasks list), but its arguments
        are inherited to next tasks.

        When we define a "@" again later, then any next tasks do not inherit previous "@" arguments (clearing)
        """

        parsed = CommandlineParsingHelper.create_grouped_arguments([
            '@', '--duration=30d',
            ':strike:start', '--general', 'now',
            '@',  # expecting: --duration will be cleared and not applied to :picket:start
            ':picket:start', '--at', 'exploiters-shop'
        ])

        self.assertEqual(
            "[Task<:strike:start (['--general', 'now', '--duration=30d'])>, " +
            "Task<:picket:start (['--at', 'exploiters-shop'])>]",
            str(parsed)
        )

    def test_global_arguments_are_changed_when_using_at_symbol_twice(self):
        """When we define a "@" task, then it is not present as a task (removed from tasks list), but its arguments
        are inherited to next tasks.

        When we define a "@" again later, then any next tasks do not inherit previous "@" arguments (clearing)
        """

        self.maxDiff = None

        parsed = CommandlineParsingHelper.create_grouped_arguments([
            '@', '--type', 'human-rights',
            ':join:activism', '--organization', 'black-lives-matter',
            '@', '--type', 'working-class-rights',
            ':join:activism', '--organization', 'international-workers-association',
            '@',
            ':send:mail'
        ])

        self.assertEqual(
            "[Task<:join:activism (['--organization', 'black-lives-matter', '--type', 'human-rights'])>, " +
            "Task<:join:activism (['--organization', 'international-workers-association', '--type', 'working-class-rights'])>, " +
            "Task<:send:mail ([])>]",
            str(parsed)
        )



#!/usr/bin/env python3

import unittest
from rkd.api.inputoutput import IO
from rkd.api.inputoutput import SystemIO
from rkd.api.inputoutput import BufferedSystemIO
from rkd.api.inputoutput import clear_formatting


class TestIO(unittest.TestCase):
    def test_is_log_level_at_least_info(self):
        """Test error level comparison

        Covers: IO.set_log_level() and IO.is_log_level_at_least()
        """

        io = IO()
        io.set_log_level('info')

        self.assertFalse(io.is_log_level_at_least('debug'))
        self.assertTrue(io.is_log_level_at_least('info'))
        self.assertTrue(io.is_log_level_at_least('warning'))
        self.assertTrue(io.is_log_level_at_least('fatal'))

    def test_set_log_level_cannot_set_invalid_log_level(self):
        """Checks validation in IO.set_log_level()"""

        io = IO()
        self.assertRaises(Exception, lambda: io.set_log_level('strikebreaker'))

    def test_inherit_silent(self):
        """Silent mode inheritance from SystemIO"""

        sys_io = SystemIO()
        sys_io.silent = True

        io = IO()
        io.inherit_silent(sys_io)

        self.assertTrue(io.is_silent())

    def test_formatting_methods_are_clearing_formatting_at_the_end(self):
        """Check that formatting methods are clearing the formatting at the end"""

        io = BufferedSystemIO()

        methods = [
            io.h1, io.h2, io.h3, io.h4, io.success_msg, io.error_msg, io.info_msg, io.print_separator, io.print_group
        ]

        for method in methods:
            try:
                method('test')
            except TypeError:
                method()

            self.assertEqual("\x1B[", io.get_value()[0:2], msg='Expected beginning of formatting')
            self.assertEqual('[0m', io.get_value().strip()[-3:], msg='Expected formatting clearing at the end')
            io.clear_buffer()

    def test_formatting_methods_are_printing_output_as_optional(self):
        """Expects that pretty-printed messages will be optional"""

        io = BufferedSystemIO()

        methods = [
            io.h1, io.h2, io.h3, io.h4, io.success_msg, io.error_msg, io.info_msg, io.print_separator, io.print_group
        ]

        for method in methods:
            self.__setattr__('is_text_optional', False)

            def opt_outln(text: str):
                self.__setattr__('is_text_optional', True)

            io.opt_outln = opt_outln

            try:
                method('test')
            except TypeError:
                method()

            self.assertTrue(self.__getattribute__('is_text_optional'),
                            msg='%s: Expected that the text will be printed through opt_outln()' % str(method))

    def test_get_log_level_raises_exception_on_unset_level(self):
        """Check DEFAULT error level and validation of not set error logging"""

        io = IO()

        self.assertEqual('info', io.get_log_level())

        io.log_level = None
        self.assertRaises(Exception, lambda: io.get_log_level())

    def test_clear_formatting_clears_simple_bash_coloring(self):
        """Test that clear_formatting() clears basic Bash coloring"""

        colored = """\x1B[93m10 June 1927 in Italy, the trial of anarchist Gino Lucetti concluded for 
attempting to assassinate Mussolini.
He was sentenced to 30 years in prison; two others received 12 years. 
He was killed by shelling in 1943 before the end of the war\x1B[0m"""

        without_coloring = clear_formatting(colored)

        self.assertFalse(without_coloring.startswith("\x1B[93m"))
        self.assertFalse(without_coloring.endswith("\x1B[0m"))




from typing import List, Callable, Union, Optional
from .argparsing import TaskArguments
from .api.syntax import TaskDeclaration, GroupDeclaration
from .context import ApplicationContext
from .exception import InterruptExecution
from .exception import TaskNotFoundException
from .aliasgroups import AliasGroup


CALLBACK_DEF = Callable[[TaskDeclaration, int, Union[GroupDeclaration, None], list], None]


class TaskResolver(object):
    """
    Responsible for finding all tasks and:
        - expanding groups (flatten tasks)
        - connecting each task to parent
        - preserve valid order of task validation/execution
    """

    _ctx: ApplicationContext
    _alias_groups: List[AliasGroup]

    def __init__(self, ctx: ApplicationContext, alias_groups: List[AliasGroup]):
        self._ctx = ctx
        self._alias_groups = alias_groups

    def resolve(self, requested_tasks: List[TaskArguments], callback: CALLBACK_DEF):
        """
        Iterate over flatten list of tasks, one by one and call a callback for each task

        :param requested_tasks:
        :param callback:
        :return:
        """

        task_num = 0

        for task_request in requested_tasks:
            task_num += 1

            try:
                self._resolve_element(task_request, callback, task_num)
            except InterruptExecution:
                return

    def _resolve_name_from_alias(self, task_name: str) -> Optional[str]:
        """Resolves task group's shortcuts eg. :hb -> :harbor"""

        for alias in self._alias_groups:
            resolved = alias.append_alias_to_task(task_name)

            if resolved:
                return resolved

    def _resolve_element(self, task_request: TaskArguments, callback: CALLBACK_DEF, task_num: int):
        """Checks task by name if it was defined in context, if yes then unpacks declarations and prepares callbacks"""

        try:
            ctx_declaration = self._ctx.find_task_by_name(task_request.name())

        # maybe a task name is an alias to other task defined by alias groups
        except TaskNotFoundException:
            task_from_alias = self._resolve_name_from_alias(task_request.name())

            if not task_from_alias:
                raise

            ctx_declaration = self._ctx.find_task_by_name(task_from_alias)

        if isinstance(ctx_declaration, TaskDeclaration):
            declarations = ctx_declaration.to_list()
            parent = None
        elif isinstance(ctx_declaration, GroupDeclaration):
            declarations = ctx_declaration.get_declarations()
            parent = ctx_declaration
        else:
            raise Exception('Cannot resolve task - unknown type "%s"' % str(ctx_declaration))

        for declaration in declarations:
            callback(
                declaration,
                task_num,
                parent,
                # the arguments there will be mixed in order:
                #  - first: defined in Makefile
                #  - second: commandline arguments
                #
                #  The argparse in Python will take the second one as priority.
                #  We do not try to remove duplications there to not increase complexity of the solution - it works now.
                declaration.get_args() + task_request.args()
            )




"""
CONTRACT (part of API)
======================

Core interfaces that should be changed WITH CAREFUL as those are parts of API.
Any breaking change there requires to bump RKD major version (see: Semantic Versioning)
"""


from tabulate import tabulate
from abc import abstractmethod, ABC as AbstractClass
from typing import Dict, List, Union, Optional
from argparse import ArgumentParser
from ..inputoutput import IO
from ..exception import UndefinedEnvironmentVariableUsageError
from ..exception import EnvironmentVariableNotUsed
from ..exception import MissingInputException
from ..exception import EnvironmentVariableNameNotAllowed
from ..taskutil import TaskUtilities
from .temp import TempManager


def env_to_switch(env_name: str) -> str:
    return '--' + env_name.replace('_', '-').lower()


class ArgumentEnv(object):
    """Represents an environment variable that should provide a value to an argparse switch

    Note: There is a list of reserved environment variables, that cannot be used. See ArgumentEnv.RESERVED_VARS
    """

    RESERVED_VARS = ['PATH', 'PWD', 'LANG', 'DISPLAY', 'SHELL', 'SHLVL', 'HOME', 'EDITOR']
    name: str
    default: str
    switch: str

    def __init__(self, name: str, switch: str = '', default: str = ''):
        self.name = name
        self.default = default
        self.switch = switch if switch else env_to_switch(name)

        self._validate()

    def _validate(self):
        if self.name in self.RESERVED_VARS:
            raise EnvironmentVariableNameNotAllowed(self.name)


class TaskDeclarationInterface(AbstractClass):
    @abstractmethod
    def to_full_name(self):
        pass

    @abstractmethod
    def get_args(self) -> List[str]:
        pass

    @abstractmethod
    def get_task_to_execute(self):  # -> TaskInterface:
        pass

    @abstractmethod
    def to_list(self) -> list:
        pass

    @abstractmethod
    def get_env(self):
        pass

    @abstractmethod
    def get_group_name(self) -> str:
        pass

    @abstractmethod
    def get_task_name(self) -> str:
        pass

    @abstractmethod
    def get_description(self) -> str:
        pass

    @abstractmethod
    def get_full_description(self) -> str:
        pass

    @abstractmethod
    def format_task_name(self, name: str) -> str:
        pass


class GroupDeclarationInterface(AbstractClass):
    @abstractmethod
    def get_declarations(self) -> Dict[str, TaskDeclarationInterface]:
        pass

    @abstractmethod
    def get_group_name(self) -> str:
        pass

    @abstractmethod
    def get_task_name(self) -> str:
        pass

    @abstractmethod
    def to_full_name(self):
        pass

    @abstractmethod
    def get_description(self) -> str:
        pass

    @abstractmethod
    def format_task_name(self, name: str) -> str:
        pass


class ContextInterface(AbstractClass):
    directories: []

    @abstractmethod
    def merge(cls, first, second):
        pass

    @abstractmethod
    def compile(self) -> None:
        pass

    def find_task_by_name(self, name: str) -> Union[TaskDeclarationInterface, GroupDeclarationInterface]:
        pass

    def find_all_tasks(self) -> Dict[str, Union[TaskDeclarationInterface, GroupDeclarationInterface]]:
        pass


class ExecutorInterface(AbstractClass):
    @abstractmethod
    def execute(self, task: TaskDeclarationInterface, task_num: int,
                parent: Union[GroupDeclarationInterface, None] = None, args: list = []):
        pass


class ExecutionContext:
    """
    Defines which objects could be accessed by Task. It's a scope of a single task execution.
    """

    declaration: TaskDeclarationInterface
    parent: Union[GroupDeclarationInterface, None]
    args: Dict[str, str]
    env: Dict[str, str]
    ctx: ContextInterface
    executor: ExecutorInterface

    # List of arguments definitions populated by Argparse (with limited parameters supported)
    # Read about "traced arguments"
    defined_args: Dict[str, dict]

    def __init__(self, declaration: TaskDeclarationInterface,
                 parent: Union[GroupDeclarationInterface, None] = None, args: Dict[str, str] = {},
                 env: Dict[str, str] = {},
                 defined_args: Dict[str, dict] = {}):
        self.declaration = declaration
        self.parent = parent
        self.args = args
        self.env = env
        self.defined_args = defined_args

    def get_env(self, name: str, switch: str = '', error_on_not_used: bool = False):
        """Get environment variable value"""
        return self.declaration.get_task_to_execute().internal_getenv(name, self.env, switch=switch,
                                                                      error_on_not_used=error_on_not_used)

    def get_arg_or_env(self, name: str) -> Union[str, None]:
        """Provides value of user input

        Usage:
            get_arg_or_env('--file-path') resolves into FILE_PATH env variable, and --file-path switch
            (file_path in argparse)

        Behavior:
            When user provided explicitly switch eg. --history-id, then it's value will be taken in priority.
            If switch --history-id was not used, but user provided HISTORY_ID environment variable,
            then it will be considered.

            If no switch provided and no environment variable provided, but a switch has
            default value - it would be returned.

            If no switch provided and no environment variable provided, the switch does not have default,
            but environment variable has a default value defined, it would be returned.

            When the --switch has default value (user does not use it, or user sets it explicitly to default value),
            and environment variable SWITCH is defined, then environment variable would be taken.

            Explicit environment variables definitions
            ------------------------------------------

            From RKD 2.1 the environment variable names can be mapped to any ArgParse switch.

            Below example maps "COMMAND" environment variable to "--cmd" switch.

            .. code:: python
                def get_declared_envs(self) -> Dict[str, Union[str, ArgumentEnv]]:
                    return {
                        'COMMAND': ArgumentEnv(name='COMMAND', switch='--cmd', default='')
                    }

        Raises:
            MissingInputException: When no switch and no environment variable was provided, then an exception is thrown.
        """
        env_name = name[2:].replace('-', '_').upper()
        env_value = None

        try:
            env_value = self.get_env(env_name, switch=name, error_on_not_used=True)
            is_env_variable_defined = True

        except EnvironmentVariableNotUsed:
            is_env_variable_defined = False

        # case 1: a --switch was used
        # case 2: --switch as default value set, environment variable is set, then pick env
        try:
            value = self.get_arg(name)

            # https://github.com/riotkit-org/riotkit-do/issues/23
            # When --switch has same value as default, and environment variable is not empty, then env has priority
            if self.defined_args[name]['default'] == value and is_env_variable_defined:
                return env_value

            if value is not None:
                return value

        except KeyError:
            pass

        # case: No --switch defined, no ENV defined
        if not is_env_variable_defined:
            raise MissingInputException(name, env_name)

        # case: No --switch defined, ENV defined
        return env_value

    def get_arg(self, name: str) -> Optional[str]:
        """Get argument or option

        Usage:
            ctx.get_arg('--name')  # for options
            ctx.get_arg('name')    # for arguments

        Raises:
            KeyError when argument/option was not defined

        Returns:
            Actual value or default value
        """

        try:
            arg_name = name[2:].replace('-', '_')

            return self.args[arg_name]
        except KeyError:
            return self.args[name]


class TaskInterface(TaskUtilities):
    _io: IO
    _ctx: ContextInterface
    _executor: ExecutorInterface
    temp: TempManager

    def internal_inject_dependencies(self, io: IO, ctx: ContextInterface,
                                     executor: ExecutorInterface, temp_manager: TempManager):
        """"""  # sphinx: skip

        self._io = io
        self._ctx = ctx
        self._executor = executor
        self.temp = temp_manager

    def copy_internal_dependencies(self, task):
        """Allows to execute a task-in-task, by copying dependent services from one task to other task
        """

        task.internal_inject_dependencies(self._io, self._ctx, self._executor, self.temp)

    @abstractmethod
    def get_name(self) -> str:
        """Task name  eg. ":sh"
        """
        pass

    @abstractmethod
    def get_group_name(self) -> str:
        """Group name where the task belongs eg. ":publishing", can be empty.
        """

        pass

    def get_become_as(self) -> str:
        """User name in UNIX/Linux system, optional.
        When defined, then current task will be executed as this user (WARNING: a forked process would be started)"""

        return ''

    def should_fork(self) -> bool:
        """Decides if task should be ran in a separate Python process (be careful with it)"""

        return self.get_become_as() != ''

    def get_description(self) -> str:
        return ''

    @abstractmethod
    def execute(self, context: ExecutionContext) -> bool:
        """ Executes a task. True/False should be returned as return """
        pass

    @abstractmethod
    def configure_argparse(self, parser: ArgumentParser):
        """ Allows a task to configure ArgumentParser (argparse) """

        pass

    def get_full_name(self):
        """ Returns task full name, including group name """

        return self.get_group_name() + self.get_name()

    def get_declared_envs(self) -> Dict[str, Union[str, ArgumentEnv]]:
        """ Dictionary of allowed envs to override: KEY -> DEFAULT VALUE """
        return {}

    def internal_normalized_get_declared_envs(self) -> Dict[str, ArgumentEnv]:
        """"""  # sphinx: ignore

        # Method used internally, supports conversion of values from primitives to ArgumentEnv
        # as developers can specify env variables in get_declared_envs() as primitives or as ArgumentEnv
        # so, there we normalize everything
        #
        # WRAPPER OVER INTERFACE METHOD: get_declared_envs() should be defined by developer

        envs = {}

        for name, value in self.get_declared_envs().items():
            if not isinstance(value, ArgumentEnv):
                value = ArgumentEnv(name=name, default=value)

            envs[name] = value

        return envs

    def internal_getenv(self, env_name: str, envs: Dict[str, str], switch: str = '',
                        error_on_not_used: bool = False) -> str:
        """"""  # sphinx: ignore

        declared_envs = self.internal_normalized_get_declared_envs()

        # find env by switch, when env was defined to be non-standard name
        if switch:
            for env in declared_envs.values():
                if env.switch == switch:
                    self.io().debug('Resolved environment "%s" from switch "%s"' % (env_name, env.switch))
                    env_name = env.name

        if env_name not in declared_envs:
            raise UndefinedEnvironmentVariableUsageError(
                'Attempt to use not declared environment variable. ' +
                'Please report the problem to the maintainers of this task, not to RKD (unless it is a core task)'
            )

        # return default value
        if env_name not in envs:
            if error_on_not_used:
                raise EnvironmentVariableNotUsed(env_name)

            return declared_envs[env_name].default

        return envs[env_name]

    def is_silent_in_observer(self) -> bool:
        """"""  # sphinx: skip
        return False

    def io(self) -> IO:
        """Gives access to Input/Output object"""

        return self._io

    def format_task_name(self, name: str) -> str:
        """Allows to add a fancy formatting to the task name, when the task is displayed eg. on the :tasks list"""

        return name

    def py(self, code: str, become: str = None, capture: bool = False, script_path: str = None) -> Union[str, None]:
        """Executes a Python code in a separate process

        NOTICE: Use instead of subprocess. Raw subprocess is less supported and output from raw subprocess
                may be not catch properly into the logs
        """

        return super().py(
            code=code, become=become, capture=capture, script_path=script_path
        )

    def sh(self, cmd: str, capture: bool = False, verbose: bool = False, strict: bool = True,
           env: dict = None) -> Union[str, None]:
        """Executes a shell script in bash. Throws exception on error.
        To capture output set capture=True

        NOTICE: Use instead of subprocess. Raw subprocess is less supported and output from raw subprocess
                may be not catch properly into the logs
        """
        return super().sh(
            cmd=cmd, capture=capture, verbose=verbose, strict=strict, env=env
        )

    def exec(self, cmd: str, capture: bool = False, background: bool = False) -> Union[str, None]:
        """Starts a process in shell. Throws exception on error.
        To capture output set capture=True

        NOTICE: Use instead of subprocess. Raw subprocess is less supported and output from raw subprocess
                may be not catch properly into the logs
        """
        return super().exec(cmd=cmd, capture=capture, background=background)

    def rkd(self, args: list, verbose: bool = False, capture: bool = False) -> str:
        """Spawns an RKD subprocess

        NOTICE: Use instead of subprocess. Raw subprocess is less supported and output from raw subprocess
                may be not catch properly into the logs
        """
        return super().rkd(args=args, verbose=verbose, capture=capture)

    def silent_sh(self, cmd: str, verbose: bool = False, strict: bool = True,
                  env: dict = None) -> bool:
        """sh() shortcut that catches errors and displays using IO().error_msg()

        NOTICE: Use instead of subprocess. Raw subprocess is less supported and output from raw subprocess
                may be not catch properly into the logs
        """
        return super().silent_sh(cmd=cmd, verbose=verbose, strict=strict, env=env)

    def __str__(self):
        return 'Task<' + self.get_full_name() + '>'

    @staticmethod
    def table(header: list, body: list, tablefmt: str = "simple",
              floatfmt: str = 'g',
              numalign: str = "decimal",
              stralign:str = "left",
              missingval: str = '',
              showindex: str = "default",
              disable_numparse: bool = False,
              colalign: str = None):

        """Renders a table

        Parameters:
            header:
            body:
            tablefmt:
            floatfmt:
            numalign:
            stralign:
            missingval:
            showindex:
            disable_numparse:
            colalign:

        Returns:
            Formatted table as string
        """

        return tabulate(body, headers=header, floatfmt=floatfmt, numalign=numalign, tablefmt=tablefmt,
                        stralign=stralign, missingval=missingval, showindex=showindex,
                        disable_numparse=disable_numparse, colalign=colalign)


class ArgparseArgument(object):
    """Represents a add_argument(*args, **kwargs)"""

    args: list
    kwargs: dict

    def __init__(self, args: list = None, kwargs: dict = None):
        if args is None:
            args = []

        if kwargs is None:
            kwargs = {}

        self.args = args
        self.kwargs = kwargs



#!/usr/bin/env python3

import sys
import os
from dotenv import load_dotenv
from .argparsing import CommandlineParsingHelper
from .context import ContextFactory, ApplicationContext
from .resolver import TaskResolver
from .validator import TaskDeclarationValidator
from .execution.executor import OneByOneTaskExecutor
from .exception import TaskNotFoundException
from .api.inputoutput import SystemIO, LEVEL_INFO as LOG_LEVEL_INFO
from .aliasgroups import parse_alias_groups_from_env


class RiotKitDoApplication:
    _ctx: ApplicationContext
    _tasks_to_execute = []

    @staticmethod
    def load_environment():
        paths = os.getenv('RKD_PATH', '').split(':')

        for path in paths:
            if os.path.isfile(path + '/.env'):
                load_dotenv(path + '/.env')

        load_dotenv(dotenv_path=os.getcwd() + '/.env')

    @staticmethod
    def prepend_development_paths():
        """Add ./src at the beginning of PYTHONPATH - very useful for development"""

        sys.path = [os.getcwd() + '/src'] + sys.path

    def main(self, argv: list):
        if not argv[1:]:
            self.print_banner_and_exit()

        # system wide IO instance with defaults, the :init task should override those settings
        io = SystemIO()
        io.silent = True
        io.log_level = LOG_LEVEL_INFO

        # load context of components
        self._ctx = ContextFactory(io).create_unified_context()

        resolver = TaskResolver(self._ctx, parse_alias_groups_from_env(os.getenv('RKD_ALIAS_GROUPS', '')))
        executor = OneByOneTaskExecutor(self._ctx)

        # iterate over each task, parse commandline arguments
        requested_tasks = CommandlineParsingHelper.create_grouped_arguments([':init'] + argv[1:])

        # validate all tasks
        resolver.resolve(requested_tasks, TaskDeclarationValidator.assert_declaration_is_valid)

        # execute all tasks
        resolver.resolve(requested_tasks, executor.execute)

        executor.get_observer().execution_finished()

        sys.exit(1 if executor.get_observer().has_at_least_one_failed_task() else 0)

    @staticmethod
    def print_banner_and_exit():
        with open(os.path.dirname(os.path.realpath(__file__)) + '/misc/banner.txt', 'rb') as banner_file:
            print(banner_file.read().replace(b'\\x1B', b'\x1B').decode('utf-8'))

        sys.exit(0)


def main():
    app = RiotKitDoApplication()
    app.prepend_development_paths()
    app.load_environment()

    try:
        app.main(argv=sys.argv)
    except TaskNotFoundException as e:
        print(e)
        sys.exit(127)


if __name__ == '__main__':
    main()




from typing import List
from typing import Optional

"""
Alias Groups
============

RKD_ALIAS_GROUPS

Allows to define aliases to groups eg. task :harbor:start can be started with just :start when alias is ':harbor->'
It is a feature that allows to create custom RKD distributions with "main tasks" shown, while the rest hidden, while the 
python package could be still imported regularly into RKD and used with :harbor:start - without having custom 
binary/distribution
"""


class AliasGroup(object):
    def __init__(self, src: str, dst: str):
        self._src = src
        self._dst = dst

    def append_alias_to_task(self, task_name) -> Optional[str]:
        if task_name[0:len(self._src)] == self._src:
            return self._dst + task_name[len(self._src):]

        return None

    def get_aliased_task_name(self, task_name) -> Optional[str]:
        if task_name[0:len(self._dst)] == self._dst:
            return self._src + task_name[len(self._dst):]

        return None


def parse_alias_groups_from_env(value: str) -> List[AliasGroup]:
    groups = value.replace(' ', '').split(',')
    parsed = []

    for group in groups:
        try:
            src, dst = group.split('->')
        except ValueError:
            continue

        parsed.append(AliasGroup(src, dst))

    return parsed



import os
from typing import Tuple
from typing import List
from uuid import uuid4
from os import unlink as delete_file
from os import chmod
from os import mkdir
from os.path import realpath


class TempManager(object):
    """
    Manages temporary files inside .rkd directory
    Using this class you make sure your code is more safe to use on Continuous Integration systems (CI)

    Usage:
        path = self.temp.assign_temporary_file(mode=0o755)

    """

    assigned: List[str]
    chdir: str

    def __init__(self, chdir: str = './.rkd/'):
        self.assigned = []
        self.chdir = chdir

    def create_tmp_file_path(self) -> Tuple[str, str]:
        tmp_vault_filename = '.tmp-' + str(uuid4())
        tmp_vault_path = self.chdir + tmp_vault_filename

        self.assigned.append(tmp_vault_path)

        return tmp_vault_path, tmp_vault_filename

    def assign_temporary_file(self, mode: int = 0o755) -> str:
        """Assign a path for writing temporary files in RKD workspace

        Note: The RKD is executing the finally_clean_up() at the end of each task

        Usage:
            try:
                path = RKDTemp.assign_temporary_file_path()
                # (...) some action there

            finally:
                RKDTemp.finally_clean_up()

        """

        if not os.path.isdir(self.chdir):
            mkdir(self.chdir)

        path = self.create_tmp_file_path()[0]

        with open(path, 'w') as f:
            f.write('')

            chmod(path=path, mode=mode)

        return realpath(path)

    def finally_clean_up(self):
        """Used to clean up all temporary files at the end of the code execution

        TaskExecutor is running this method after each finished task
        """

        for path in self.assigned:
            try:
                delete_file(path)
            except Exception:
                pass

        self.assigned = []
